from __future__ import annotations
from typing import Callable
from ehelply_bootstrapper.integrations.integration import Integration
from ehelply_bootstrapper.utils.service import ServiceMeta
from ehelply_microservice_library.integrations.fact import get_fact_endpoint
from ehelply_microservice_library.integrations.logging import Log

from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.drivers.fast_api_utils import responses

from fastapi import HTTPException
from pydantic import BaseModel
from fastapi.encoders import jsonable_encoder
import json
from ehelply_microservice_library.utils.ehelply_cloud import (
    ehelply_cloud_access,
    CloudParticipantAuthRequest,
    CloudParticipantResponse,
    KeysDB,
    ProjectDB,
    ProjectMembersDB,
    ProjectCredentialDB,
    ProjectMemberRoleEnum,
    ProjectStatus,
    is_allowed
)


class UeriModel(BaseModel):
    company: str
    product: str
    logical_group: str
    location: str
    service: str
    service_type: str
    type_id: str
    drill_segments: list[str]


class Ueri(Integration):
    """
    Monitor integration is used to talk to the ehelply-meta microservice
    """

    def __init__(
            self, service_meta: ServiceMeta, handlers: dict[str, Callable[[str], BaseModel | dict | list]],
            custom_drill: Callable[[list[str], BaseModel | dict | list], any] | None = None,
    ) -> None:
        super().__init__("ueri")
        self.service_meta = service_meta
        self.handlers = handlers
        self.custom_drill = custom_drill

    def load(self):
        pass

    def parse_ueri(self, ueri: str, exception_if_invalid: bool = True) -> UeriModel | None:
        ueri_components: list[str] = ueri.split(":")
        if len(ueri_components) < 8 or len(ueri_components) > 9:
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400, detail="Invalid ueri"
                )
            else:
                return None

        if ueri_components[0] != "ueri":
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400, detail="Invalid ueri format"
                )
            else:
                return None

        drill_segments: list[str] = []
        if len(ueri_components) == 9:
            drill_segments = ueri_components[8].split("/")

        ueri_model: UeriModel = UeriModel(
            company=ueri_components[1],
            product=ueri_components[2],
            logical_group=ueri_components[3],
            location=ueri_components[4],
            service=ueri_components[5],
            service_type=ueri_components[6],
            type_id=ueri_components[7],
            drill_segments=drill_segments
        )

        if ueri_model.company != "ehelply":
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400, detail="This ueri endpoint can only handle company values of: ehelply"
                )
            else:
                return None
        if ueri_model.product != "superstack":
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400, detail="This ueri endpoint can only handle product values of: superstack"
                )
            else:
                return None
        if ueri_model.location != "ca-01":
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400, detail="This ueri endpoint can only handle location values of: ca-01"
                )
            else:
                return None
        if ueri_model.service != self.service_meta.key:
            if exception_if_invalid:
                raise HTTPException(
                    status_code=400,
                    detail=f"This ueri endpoint can only handle service values of: {self.service_meta.key}"
                )
            else:
                return None

        return ueri_model

    def handle(self, cloud_participant: CloudParticipantResponse, ueri: UeriModel) -> BaseModel | dict | list:
        is_allowed(
            cloud_participant=cloud_participant,
            project_uuid=ueri.logical_group,
            # allow_privileged=True
        )

        if ueri.service_type not in self.handlers:
            raise HTTPException(
                status_code=400,
                detail=f"This ueri endpoint can only handle service type values of: {self.handlers.keys()}"
            )

        return self.handlers[ueri.service_type](ueri.type_id)

    def __default_drill(self, drill_segments: list[str], result: BaseModel | dict | list) -> any:
        for segment in drill_segments:
            if segment.isdigit():
                segment = int(segment)
            result = result[segment]
        return result

    def drill(self, ueri: UeriModel, result: BaseModel | dict | list) -> any:
        try:
            drill = self.__default_drill
            if self.custom_drill:
                drill = self.custom_drill
            return drill(ueri.drill_segments, result)
        except:
            raise HTTPException(
                status_code=400,
                detail=f"Invalid drill segments"
            )



import typer
from pathlib import Path
from typing import Optional
from ehelply_bootstrapper.utils.state import State
from ehelply_microservice_library.lightweight_bootstrap import lightweight_bootstrap
from ehelply_microservice_library.utils.secret_names import secret_name_updater_vault
from ehelply_bootstrapper.drivers.aws_utils.aws_asm import ASM
import os

cli = typer.Typer()


@cli.command()
def install(
        root: Optional[Path] = typer.Option(None, help="Path to where the ehelply-package.json file can be found. If blank, default is current working directory")
):
    """
    Install the latest service template updates
    """
    with typer.progressbar(length=4, label="Updating") as progress:
        root_path: Path = root
        if not root_path:
            root_path = Path(os.getcwd())
        # root_path: Path = Path(Path(__file__).resolve().parents[3])

        lightweight_bootstrap(root_path)
        progress.update(1)

        secret = ASM.get_secret(secret_name_updater_vault())
        access_key: str = secret.access
        secret_key: str = secret.secret
        api_base_url: str = State.config.updater.endpoint

        from ehelply_updater.updater import Updater, Config

        config: Config = Config(
            api_base_url=api_base_url + "/templates",
            update_dir=str(root_path),
            package_file=str(root_path.joinpath("ehelply-package.json")),
            access_key=access_key,
            secret_key=secret_key,
        )
        updater: Updater = Updater(config=config)
        progress.update(1)

        update_available: bool = updater.is_update_required()
        progress.update(1)

        if update_available:
            updater.pp_upgrade_info(updater.update())
        else:
            typer.echo("No service template updates available")

        progress.update(1)


@cli.command()
def preview(
        root: Optional[Path] = typer.Option(None, help="Path to where the ehelply-package.json file can be found. If blank, default is current working directory")
):
    """
    Preview the latest service template updates. Won't install, but will show the changelog and changes
    """
    with typer.progressbar(length=4, label="Processing") as progress:
        root_path: Path = root
        if not root_path:
            root_path = Path(os.getcwd())
        #root_path: Path = Path(Path(__file__).resolve().parents[3])

        lightweight_bootstrap(root_path)
        progress.update(1)

        secret = ASM.get_secret(secret_name_updater_vault())
        access_key: str = secret.access
        secret_key: str = secret.secret
        api_base_url: str = State.config.updater.endpoint

        from ehelply_updater.updater import Updater, Config


        config: Config = Config(
            api_base_url=api_base_url + "/templates",
            update_dir=str(root_path),
            package_file=str(root_path.joinpath("ehelply-package.json")),
            access_key=access_key,
            secret_key=secret_key,
        )
        updater: Updater = Updater(config=config)
        progress.update(1)

        update_available: bool = updater.is_update_required()
        progress.update(1)

        if update_available:
            updater.pp_upgrade_info(updater.preview())
        else:
            typer.echo("No service template updates available")

        progress.update(1)


@cli.command()
def review(
        start_version: str = typer.Option(..., prompt=True, help="Start version"),
        end_version: str = typer.Option(..., prompt=True, help="End version"),
        root: Optional[Path] = typer.Option(None, help="Path to where the ehelply-package.json file can be found. If blank, default is current working directory")
):
    """
    Compare the changes and changelog between two versions
    """
    with typer.progressbar(length=4, label="Processing") as progress:
        root_path: Path = root
        if not root_path:
            root_path = Path(os.getcwd())
        #root_path: Path = Path(Path(__file__).resolve().parents[3])

        lightweight_bootstrap(root_path)
        progress.update(1)

        secret = ASM.get_secret(secret_name_updater_vault())
        access_key: str = secret.access
        secret_key: str = secret.secret
        api_base_url: str = State.config.updater.endpoint

        from ehelply_updater.updater import Updater, Config, UpdateInfo

        root_path: Path = Path(Path(__file__).resolve().parents[3])
        config: Config = Config(
            api_base_url=api_base_url + "/templates",
            update_dir=str(root_path),
            package_file=str(root_path.joinpath("ehelply-package.json")),
            access_key=access_key,
            secret_key=secret_key,
        )
        updater: Updater = Updater(config=config)
        progress.update(1)

        import requests

        response = requests.get(
            updater.config.api_base_url + "/versions/" + start_version + "/updates/" + end_version,
            headers=updater.headers)

        if response.status_code == 200:
            updater.pp_upgrade_info(UpdateInfo(**response.json()))
        else:
            raise typer.echo("Unable to get update information")

        progress.update(1)


from __future__ import annotations
from ehelply_bootstrapper.integrations.integration import Integration
from ehelply_microservice_library.integrations.fact import get_fact_endpoint
from ehelply_microservice_library.integrations.logging import Log

from ehelply_bootstrapper.utils.state import State

from pydantic import BaseModel
from fastapi.encoders import jsonable_encoder
import json


class UsageType(BaseModel):
    """
    Usage_type_key: The string usage type key name
    Scaling: The default amount of scaling to apply to a particular usage type. This will help remove the constant
        need for *1000000 for usage types that dont need that level of precision
    """
    usage_type_key: str
    scaling: int = 1


class UsageMQ(BaseModel):
    project_uuid: str
    usage_key: str
    quantity: int  # Quantity formats represented by a x10000000 integer. Precision to the millonth


class Monitor(Integration):
    """
    Monitor integration is used to talk to the ehelply-meta microservice
    """

    def __init__(self, supported_usage_types: list[UsageType]) -> None:
        super().__init__("monitor")

        self.sqs = State.aws.make_client("sqs")
        self.supported_usage_types: dict[str, int] = {
            item.usage_type_key: item.scaling for item in supported_usage_types
        }

    def load(self):
        pass

    def add_usage(self, usage: UsageMQ) -> bool:
        """
        Add usage to a project

        NOTE: Quantity formats represented by a x10000000 integer. Precision to the millonth

        :return:
        """
        if usage.usage_key not in self.supported_usage_types:
            State.logger.warning(
                f"Usage type has not been registered. Override get_usage_types() in service.py and add {usage.usage_key} there."
            )
        else:
            usage.quantity *= self.supported_usage_types[usage.usage_key]

        if usage.project_uuid == 'ehelply-resources' or usage.project_uuid == 'ehelply-cloud':
            return False

        self.sqs.send_message(
            QueueUrl=get_fact_endpoint("mq-usage"),
            MessageBody=json.dumps(jsonable_encoder(usage.dict()))
        )

        return True



from ehelply_bootstrapper.integrations.integration import Integration
import threading
from ehelply_logger.Logger import Logger

import requests
import time

from typing import List

from pydantic import BaseModel
from ehelply_bootstrapper.utils.environment import Environment

from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.utils.cryptography import Encryption


class FactBase(BaseModel):
    name: str
    data: dict


def get_fact(fact_name: str):
    from ehelply_bootstrapper.utils.state import State
    fact: Fact = State.integrations.get("fact")
    return fact.facts[fact_name]


def get_fact_stage(fact_name: str):
    print(State.integrations.get("fact").facts)
    return get_fact(fact_name)


def get_fact_endpoint(fact_name: str):
    return get_fact(fact_name)['endpoint']


class Fact(Integration):
    def __init__(self, service_gatekeeper_key: str, logger: Logger, fact_names: List[str] = None) -> None:
        super().__init__("fact")
        self.fact_names = []
        if fact_names:
            self.fact_names = fact_names
        self.facts = {}
        self.logger = logger.spinoff()
        self.thread: FactThread = None
        self.fact_query_delay_minutes: int = 20
        self.enc: Encryption = Encryption([service_gatekeeper_key.encode(Encryption.STRING_ENCODING)])
        self.requests: requests.Session = requests.Session()

    def init(self):
        secret_token: str = State.config.m2m.auth.secret_key
        access_token: str = State.config.m2m.auth.access_key

        if len(secret_token) == 0 or len(access_token) == 0:
            State.logger.warning("SDK (M2M) credentials are not set. Check the m2m.yaml config")

        secret_token = self.enc.decrypt_str(secret_token.encode(Encryption.STRING_ENCODING))
        access_token = self.enc.decrypt_str(access_token.encode(Encryption.STRING_ENCODING))
        self.requests.headers.update({
            'X-Secret-Token': secret_token,
            'X-Access-Token': access_token,
            'Ehelply-Project': 'ehelply-cloud'
        })

        # Initial load to prevent race conditions
        self._query()
        self.thread = FactThread(self).start()

    def load(self):
        pass

    def _query(self):
        State.logger.info("  -> Fetching Latest Facts")
        for fact_name in self.fact_names:
            State.logger.debug("     * " + State.config.facts.endpoint + "/facts/" + fact_name)
            response = self.requests.get(State.config.facts.endpoint + "/facts/" + fact_name)
            try:
                if response.status_code == 200:
                    self.facts[fact_name] = response.json()['data']
                else:
                    State.logger.warning(f"Fact not found: {fact_name}")
            except:
                raise Exception(
                    "Communication with ehelply-facts failed. There are several possible reasons for this: incorrect fact name or you're sending the request from an IP address that has not been whitelisted.")

    def refresh(self):
        self._query()

    def run(self):
        while True:
            time.sleep(self.fact_query_delay_minutes * 60)
            self._query()

    def push_fact(self, fact: FactBase):
        response = self.requests.post(State.config.facts.endpoint + "/facts", json={"fact": fact.dict()})
        return response


class FactThread(threading.Thread):
    def __init__(self, fact: Fact):
        super().__init__()
        self.fact: Fact = fact
        self.daemon = True

    def run(self) -> None:
        self.fact.run()



"""
3rd Party Libraries
Not much interesting here. We have some default imports for FastAPI endpoints and common SQL Alchemy use cases
"""
from fastapi import APIRouter, Body, Depends

"""
eHelply Bootstrapper
Now we are getting interesting.
- State provides you with access to the app state
- get_db is used with Dependency injection to obtain a fresh DB connection session
    `db: Session = Depends(get_db)`
- Responses imports a bunch of default HTTP responses that eHelply has standardized on. Use these where possible as your
    return from each endpoint
"""
from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.drivers.fast_api_utils.responses import *

"""
Service Template Integration Dependency Injection
Super duper interesting!
This imports ALL of the integration dependency injection functions. Examples of how these are used are shown in the 
  endpoints below
"""
from ehelply_microservice_library.integrations.router_integrations import *
from ehelply_microservice_library.integrations.ueri import Ueri, UeriModel
from ehelply_microservice_library.utils.ehelply_cloud import (
    ehelply_cloud_access,
    CloudParticipantAuthRequest,
    CloudParticipantResponse,
)

"""
Router
This is a self-contained router instance. By itself it does nothing, but by attaching endpoints to it using decorator
  notation, you can make an epic router!
"""
router = APIRouter()


@router.post(
    '/db/migrate',
    tags=["db"],
)
async def db_migrate():
    """
    Runs DB migrations
    :return:
    """
    try:
        State.logger.info("Running Migrations..")
        from alembic.config import Config
        import alembic.command

        config = Config('alembic.ini')
        config.attributes['configure_logger'] = False

        alembic.command.upgrade(config, 'head')
    except:
        return http_400_bad_request(error_message="Migrations failed")

    return http_200_ok({"message": "Migrations complete"})


@router.post(
    '/db/migrate/create',
    tags=["db"],
)
async def db_create_migration(data: dict = Body({})):
    """
    Runs DB migrations
    :return:
    """
    try:
        State.logger.info("Creating migration..")
        from alembic.config import Config
        import alembic.command

        config = Config('alembic.ini')
        config.attributes['configure_logger'] = False

        alembic.command.revision(config, autogenerate=True, message=data['message'])
    except:
        return http_400_bad_request(error_message="Creating migration failed")

    return http_200_ok({"message": "Creating migration complete"})


@router.post(
    '/install',
    tags=["service"],
)
async def service_install(integrations: Integrations = Depends(get_integrations), data: dict = Body({})):
    """
    Runs the microservice installer
    :return:
    """
    result = await State.bootstrapper.install(integrations, data)
    if result:
        response = {"message": "Installation complete"}
        if not isinstance(result, bool) and result is not None:
            response['result'] = result
        return http_200_ok(response)

    return http_500_error(error_message="Installation failed")


@router.post(
    '/seed',
    tags=["service"],
)
async def service_seed(integrations: Integrations = Depends(get_integrations), data: dict = Body({})):
    """
    Runs the microservice seeder
    :return:
    """
    result = await State.bootstrapper.seed(integrations, data)
    if result:
        response = {"message": "Seeding complete"}
        if not isinstance(result, bool) and result is not None:
            response['result'] = result
        return http_200_ok(response)

    return http_500_error(error_message="Seeding failed")


@router.get(
    "/ueri/{ueri}",
    tags=["service"]
)
async def get_ueri(
        ueri: str,
        auth: CloudParticipantAuthRequest = Depends(get_auth),
        ueri_integration: Ueri = Depends(get_ueri)
):
    cloud_participant: CloudParticipantResponse = await ehelply_cloud_access(
        auth=auth
    )
    ueri_model = ueri_integration.parse_ueri(ueri)
    handled_ueri = ueri_integration.handle(cloud_participant, ueri_model)
    return ueri_integration.drill(ueri_model, handled_ueri)
